# -*- coding: utf-8 -*-

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."

from aiida.tools.dbexporters.tcod_plugins import BaseTcodtranslator

class PwTcodtranslator(BaseTcodtranslator):
    """
    Quantum ESPRESSO's PW-specific input and output parameter translator
    to TCOD CIF dictionary tags.
    """
    _plugin_type_string = "quantumespresso.pw.PwCalculation"

    _smearing_aliases = {
        'gaussian': 'Gaussian',
        'gauss': 'Gaussian',
        'methfessel-paxton': 'Methfessel-Paxton',
        'm-p': 'Methfessel-Paxton',
        'mp': 'Methfessel-Paxton',
        'marzari-vanderbilt': 'Marzari-Vanderbilt',
        'cold': 'Marzari-Vanderbilt',
        'm-v': 'Marzari-Vanderbilt',
        'mv': 'Marzari-Vanderbilt',
        'fermi-dirac': 'Marzari-Vanderbilt',
        'f-d': 'Marzari-Vanderbilt',
        'fd': 'Marzari-Vanderbilt',
    }

    @classmethod
    def get_software_package(cls,calc,**kwargs):
        """
        Returns the package or program name that was used to produce
        the structure. Only package or program name should be used,
        e.g. 'VASP', 'psi3', 'Abinit', etc.
        """
        return 'Quantum ESPRESSO'

    @classmethod
    def _get_pw_energy_value(cls,calc,energy_type,**kwargs):
        """
        Returns the energy of defined type in eV.
        """
        parameters = calc.out.output_parameters
        if energy_type not in parameters.attrs():
            return None
        if energy_type + '_units' not in parameters.attrs():
            raise ValueError("energy units for {} are "
                             "unknown".format(energy_type))
        if parameters.get_attr(energy_type + '_units') != 'eV':
            raise ValueError("energy units for {} are {} "
                             "instead of eV -- unit "
                             "conversion is not possible "
                             "yet".format(energy_type,
                                          parameters.get_attr(energy_type + '_units')))
        return parameters.get_attr(energy_type)

    @classmethod
    def _get_atom_site_residual_force_Cartesian(cls,calc,index,**kwargs):
        """
        Returns an array with residual force components along the Cartesian
        axes.
        """
        try:
            array = calc.out.output_array
            return [x[index] for x in array.get_array('forces').tolist()[-1]]
        except KeyError:
            return None

    @classmethod
    def _get_BZ_integration_grid(cls,calc,**kwargs):
        """
        Returns an array with Brillouin zone point counts along each
        vector of reciprocal lattice.
        """
        try:
            array,_ = calc.inp.kpoints.get_kpoints_mesh()
            return array
        except AttributeError:
            return None
        except KeyError:
            return None

    @classmethod
    def _get_BZ_integration_grid_shift(cls,calc,**kwargs):
        """
        Returns an array with Brillouin zone point shifts along each
        vector of reciprocal lattice.
        """
        try:
            _,array = calc.inp.kpoints.get_kpoints_mesh()
            return array
        except AttributeError:
            return None
        except KeyError:
            return None

    @classmethod
    def _get_raw_integration_smearing_method(cls,calc,**kwargs):
        """
        Returns the smearing method name as string, as specified in the
        input parameters (if specified). If not 'smearing' is not
        specified, but 'occupations' == 'smearing', string with default
        value 'gaussian' is returned, as specified in
        http://www.quantum-espresso.org/wp-content/uploads/Doc/INPUT_PW.html
        """
        parameters = calc.inp.parameters
        smearing = None
        try:
            smearing = parameters.get_dict()['SYSTEM']['smearing']
        except KeyError:
            pass
        if smearing is None:
            try:
                if parameters.get_dict()['SYSTEM']['occupations'] == 'smearing':
                    smearing = 'gaussian'
            except KeyError as e:
                pass
        return smearing
        
    @classmethod
    def get_total_energy(cls,calc,**kwargs):
        """
        Returns the total energy in eV.
        """
        return cls._get_pw_energy_value(calc,'energy')

    @classmethod
    def get_one_electron_energy(cls,calc,**kwargs):
        """
        Returns one electron energy in eV.
        """
        return cls._get_pw_energy_value(calc,'energy_one_electron')

    @classmethod
    def get_exchange_correlation_energy(cls,calc,**kwargs):
        """
        Returns exchange correlation (XC) energy in eV.
        """
        return cls._get_pw_energy_value(calc,'energy_xc')

    @classmethod
    def get_ewald_energy(cls,calc,**kwargs):
        """
        Returns Ewald energy in eV.
        """
        return cls._get_pw_energy_value(calc,'energy_ewald')

    @classmethod
    def get_hartree_energy(cls,calc,**kwargs):
        """
        Returns Hartree energy in eV.
        """
        return cls._get_pw_energy_value(calc,'energy_hartree')

    @classmethod
    def get_fermi_energy(cls,calc,**kwargs):
        """
        Returns Fermi energy in eV.
        """
        return cls._get_pw_energy_value(calc,'fermi_energy')

    @classmethod
    def get_number_of_electrons(cls,calc,**kwargs):
        """
        Returns the number of electrons.
        """
        parameters = calc.out.output_parameters
        if 'number_of_electrons' not in parameters.attrs():
            return None
        return parameters.get_attr('number_of_electrons')

    @classmethod
    def get_computation_wallclock_time(cls,calc,**kwargs):
        """
        Returns the computation wallclock time in seconds.
        """
        parameters = calc.out.output_parameters
        if 'wall_time_seconds' not in parameters.attrs():
            return None
        return parameters.get_attr('wall_time_seconds')

    @classmethod
    def get_atom_site_residual_force_Cartesian_x(cls,calc,**kwargs):
        """
        Returns a list of x components for Cartesian coordinates of
        residual force for atom. The list order MUST be the same as in
        the resulting structure.
        """
        return cls._get_atom_site_residual_force_Cartesian(calc,0)

    @classmethod
    def get_atom_site_residual_force_Cartesian_y(cls,calc,**kwargs):
        """
        Returns a list of y components for Cartesian coordinates of
        residual force for atom. The list order MUST be the same as in
        the resulting structure.
        """
        return cls._get_atom_site_residual_force_Cartesian(calc,1)

    @classmethod
    def get_atom_site_residual_force_Cartesian_z(cls,calc,**kwargs):
        """
        Returns a list of z components for Cartesian coordinates of
        residual force for atom. The list order MUST be the same as in
        the resulting structure.
        """
        return cls._get_atom_site_residual_force_Cartesian(calc,2)

    @classmethod
    def get_BZ_integration_grid_X(cls,calc,**kwargs):
        """
        Returns a number of points in the Brillouin zone along reciprocal
        lattice vector X.
        """
        array = cls._get_BZ_integration_grid(calc,**kwargs)
        if array is not None:
            return array[0]
        else:
            return None

    @classmethod
    def get_BZ_integration_grid_Y(cls,calc,**kwargs):
        """
        Returns a number of points in the Brillouin zone along reciprocal
        lattice vector Y.
        """
        array = cls._get_BZ_integration_grid(calc,**kwargs)
        if array is not None:
            return array[1]
        else:
            return None

    @classmethod
    def get_BZ_integration_grid_Z(cls,calc,**kwargs):
        """
        Returns a number of points in the Brillouin zone along reciprocal
        lattice vector Z.
        """
        array = cls._get_BZ_integration_grid(calc,**kwargs)
        if array is not None:
            return array[2]
        else:
            return None

    @classmethod
    def get_BZ_integration_grid_shift_X(cls,calc,**kwargs):
        """
        Returns the shift of the Brillouin zone points along reciprocal
        lattice vector X.
        """
        array = cls._get_BZ_integration_grid_shift(calc,**kwargs)
        if array is not None:
            return array[0]
        else:
            return None

    @classmethod
    def get_BZ_integration_grid_shift_Y(cls,calc,**kwargs):
        """
        Returns the shift of the Brillouin zone points along reciprocal
        lattice vector Y.
        """
        array = cls._get_BZ_integration_grid_shift(calc,**kwargs)
        if array is not None:
            return array[1]
        else:
            return None

    @classmethod
    def get_BZ_integration_grid_shift_Z(cls,calc,**kwargs):
        """
        Returns the shift of the Brillouin zone points along reciprocal
        lattice vector Z.
        """
        array = cls._get_BZ_integration_grid_shift(calc,**kwargs)
        if array is not None:
            return array[2]
        else:
            return None

    @classmethod
    def get_integration_smearing_method(cls,calc,**kwargs):
        """
        Returns the smearing method name as string.
        """
        smearing = cls._get_raw_integration_smearing_method(calc,**kwargs)
        if smearing is None:
            return None
        elif smearing in cls._smearing_aliases:
            return cls._smearing_aliases[smearing]
        else:
            return 'other'

    @classmethod
    def get_integration_smearing_method_other(cls,calc,**kwargs):
        """
        Returns the smearing method name as string if the name is different
        from specified in cif_dft.dic.
        """
        smearing = cls._get_raw_integration_smearing_method(calc,**kwargs)
        if smearing is None or smearing in cls._smearing_aliases:
            return None
        else:
            return smearing

    @classmethod
    def get_integration_Methfessel_Paxton_order(cls,calc,**kwargs):
        """
        Returns the order of Methfessel-Paxton approximation if used.
        """
        if cls.get_integration_smearing_method(calc,**kwargs) == \
           'Methfessel-Paxton':
            return 1
        else:
            return None

    @classmethod
    def get_kinetic_energy_cutoff_wavefunctions(cls,calc,**kwargs):
        """
        Returns kinetic energy cutoff for wavefunctions in eV.
        """
        from aiida.common.constants import ry_to_ev
        parameters = calc.inp.parameters
        ecutwfc = None
        try:
            ecutwfc = parameters.get_dict()['SYSTEM']['ecutwfc']
        except KeyError:
            pass
        if ecutwfc is None:
            return None
        else:
            return ecutwfc * ry_to_ev

    @classmethod
    def get_kinetic_energy_cutoff_charge_density(cls,calc,**kwargs):
        """
        Returns kinetic energy cutoff for charge density in eV.

        .. note :: by default returns 4 * ecutwfc, as indicated in
            http://www.quantum-espresso.org/wp-content/uploads/Doc/INPUT_PW.html
        """
        from aiida.common.constants import ry_to_ev
        parameters = calc.inp.parameters
        try:
            return parameters.get_dict()['SYSTEM']['ecutrho'] * ry_to_ev
        except KeyError:
            pass
        ecutwfc = cls.get_kinetic_energy_cutoff_wavefunctions(calc)
        if ecutwfc is None:
            return None
        else:
            return 4 * ecutwfc

    @classmethod
    def get_kinetic_energy_cutoff_EEX(cls,calc,**kwargs):
        """
        Returns kinetic energy cutoff for exact exchange (EEX)
        operator in eV.

        .. note :: by default returns ecutrho, as indicated in
            http://www.quantum-espresso.org/wp-content/uploads/Doc/INPUT_PW.html
        """
        from aiida.common.constants import ry_to_ev
        parameters = calc.inp.parameters
        try:
            return parameters.get_dict()['SYSTEM']['ecutfock'] * ry_to_ev
        except KeyError:
            pass
        return cls.get_kinetic_energy_cutoff_charge_density(calc)



# -*- coding: utf-8 -*-

import json

from sqlalchemy.schema import Column
from sqlalchemy.types import Integer, DateTime, String, Text
from sqlalchemy.dialects.postgresql import JSONB

from aiida.utils import timezone
from aiida.backends.sqlalchemy.models.base import Base
from aiida.common.exceptions import ValidationError

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__authors__ = "The AiiDA team."
__version__ = "0.7.0.1"


class DbLog(Base):
    __tablename__ = "db_dblog"

    id = Column(Integer, primary_key=True)

    time = Column(DateTime(timezone=True), default=timezone.now)
    loggername = Column(String(255), index=True)
    levelname = Column(String(255), index=True)

    objname = Column(String(255), index=True)
    objpk = Column(Integer, index=True, nullable=True)

    message = Column(Text(), nullable=True)
    _metadata = Column('metadata', JSONB)

    def __init__(self, loggername="", levelname="", objname="", objpk=None,
                 message=None, metadata=None):

        if not loggername or not levelname:
            raise ValidationError(
                "The loggername and levelname can't be empty")

        self.loggername = loggername
        self.levelname = levelname
        self.objname = objname
        self.objpk = objpk
        self.message = message
        self._metadata = metadata or {}


    def __str__(self):
        return "DbComment for [{} {}] on {}".format(
            self.dbnode.get_simple_name(),
            self.dbnode.id, timezone.localtime(self.ctime).strftime("%Y-%m-%d")
        )

    @classmethod
    def add_from_logrecord(cls, record):
        """
        Add a new entry from a LogRecord (from the standard python
        logging facility). No exceptions are managed here.
        """

        objpk = record.__dict__.get('objpk', None)
        objname = record.__dict__.get('objname', None)

        # Filter: Do not store in DB if no objpk and objname is given
        if objpk is None or objname is None:
            return

        new_entry = cls(loggername=record.name,
                        levelname=record.levelname,
                        objname=objname,
                        objpk=objpk,
                        message=record.getMessage(),
                        metadata=record.__dict__)
        new_entry.save()



# -*- coding: utf-8 -*-

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__authors__ = "The AiiDA team."
__version__ = "0.7.0.1"

__all__ = ['delete_computer', 'django_filter', 'get_attr']

from sqlalchemy import inspect
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm.mapper import Mapper
from sqlalchemy.types import Integer, Boolean

from aiida.backends import sqlalchemy as sa
from aiida.common.exceptions import InvalidOperation
from aiida.orm.implementation.sqlalchemy.computer import Computer


def delete_computer(computer):
    """
    Delete a computer from the DB.
    It assumes that the DB backend does the proper checks and avoids
    to delete computers that have nodes attached to them.

    Implemented as a function on purpose, otherwise complicated logic would be
    needed to set the internal state of the object after calling
    computer.delete().
    """
    if not isinstance(computer, Computer):
        raise TypeError("computer must be an instance of "
                        "aiida.orm.computer.Computer")

    try:
        sa.session.delete(computer.dbcomputer)
        sa.session.commit()
    except SQLAlchemyError:
        raise InvalidOperation("Unable to delete the requested computer: there"
                               "is at least one node using this computer")

def iter_dict(attrs):
    if isinstance(attrs, dict):
        for key in sorted(attrs.iterkeys()):
            it = iter_dict(attrs[key])
            for k, v in it:
                new_key = key
                if k:
                    new_key += "." + str(k)
                yield new_key, v
    elif isinstance(attrs, list):
        for i, val in enumerate(attrs):
            it = iter_dict(val)
            for k, v  in it:
                new_key = str(i)
                if k:
                    new_key += "." + str(k)
                yield new_key, v
    else:
        yield "", attrs


def get_attr(attrs, key):
    path = key.split('.')

    d = attrs
    for p in path:
        if p.isdigit():
            p = int(p)
        # Let it raise the appropriate exception
        d = d[p]

    return d


def _create_op_func(op):
    def f(attr, val):
        return getattr(attr, op)(val)

    return f


_from_op = {
    'in': _create_op_func('in_'),
    'gte': _create_op_func('__ge__'),
    'gt': _create_op_func('__gt__'),
    'lte': _create_op_func('__le__'),
    'lt': _create_op_func('__lt__'),
    'eq': _create_op_func('__eq__'),
    'startswith': lambda attr, val: attr.like('{}%'.format(val)),
    'contains': lambda attr, val: attr.like('%{}%'.format(val)),
    'endswith': lambda attr, val: attr.like('%{}'.format(val)),
    'istartswith': lambda attr, val: attr.ilike('{}%'.format(val)),
    'icontains': lambda attr, val: attr.ilike('%{}%'.format(val)),
    'iendswith': lambda attr, val: attr.ilike('%{}'.format(val))
}


def django_filter(cls_query, **kwargs):
    # Pass the query object you want to use.
    # This also assume a AND between each arguments

    cls = inspect(cls_query)._entity_zero().type
    q = cls_query

    # We regroup all the filter on a relationship at the same place, so that
    # when a join is done, we can filter it, and then reset to the original
    # query.
    current_join = None


    tmp_attr = dict(key=None, val=None)
    tmp_extra = dict(key=None, val=None)

    for key in sorted(kwargs.iterkeys()):
        val = kwargs[key]

        join, field, op = [None]*3

        splits = key.split("__")
        if len(splits) > 3:
            raise ValueError("Too many parameters to handle.")
        # something like "computer__id__in"
        elif len(splits) == 3:
            join, field, op = splits
        # we have either "computer__id", which means join + field quality or
        # "id__gte" which means field + op
        elif len(splits) == 2:
            if splits[1] in _from_op.iterkeys():
                field, op = splits
            else:
                join, field = splits
        else:
            field = splits[0]

        if "dbattributes" == join:
            if "val" in field:
                field = "val"
            if field in ["key", "val"]:
                tmp_attr[field] = val
            continue
        elif "dbextras" == join:
            if "val" in field:
                field = "val"
            if field in ["key", "val"]:
                tmp_extra[field] = val
            continue

        current_cls = cls
        if join:
            if current_join != join:
                q = q.join(join, aliased=True)
                current_join = join

            current_cls = filter(lambda r: r[0] == join,
                                 inspect(cls).relationships.items()
                                 )[0][1].argument
            if isinstance(current_cls, Mapper):
                current_cls = current_cls.class_
            else:
                current_cls = current_cls()

        else:
            if current_join is not None:
                # Filter on the queried class again
                q = q.reset_joinpoint()
                current_join = None

        if field == "pk":
            field = "id"

        filtered_field = getattr(current_cls, field)
        if not op:
            op = "eq"
        f = _from_op[op]

        q = q.filter(f(filtered_field, val))

    # We reset one last time
    q.reset_joinpoint()

    key = tmp_attr["key"]
    if key:
        val = tmp_attr["val"]
        if val:
            q = q.filter(apply_json_cast(cls.attributes[key], val) == val)
        else:
            q = q.filter(cls.attributes.has_key(tmp_attr["key"]))
    key = tmp_extra["key"]
    if key:
        val = tmp_extra["val"]
        if val:
            q = q.filter(apply_json_cast(cls.extras[key], val) == val)
        else:
            q = q.filter(cls.extras.has_key(tmp_extra["key"]))

    return q


def apply_json_cast(attr, val):
    if isinstance(val, basestring):
        attr = attr.astext
    if isinstance(val, int) or isinstance(val, long):
        attr = attr.astext.cast(Integer)
    if isinstance(val, bool):
        attr = attr.astext.cast(Boolean)

    return attr



# -*- coding: utf-8 -*-
from aiida.backends.utils import load_dbenv, is_dbenv_loaded

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__authors__ = "The AiiDA team."
__version__ = "0.7.0.1"

if not is_dbenv_loaded():
    load_dbenv()

from aiida.orm import load_node
from aiida.orm.utils import DataFactory
from aiida.workflows2.db_types import Float, Str, NumericType, SimpleData
from aiida.orm.code import Code
from aiida.orm.data.structure import StructureData
from aiida.workflows2.run import run, asyncd
from aiida.workflows2.fragmented_wf import (FragmentedWorkfunction,
                                            ResultToContext)
from common import generate_scf_input_params
from examples.workflows2.diamond_fcc import rescale
from aiida.orm.calculation.job.quantumespresso.pw import PwCalculation


# Set up the factories
ParameterData = DataFactory("parameter")
KpointsData = DataFactory("array.kpoints")


PwProcess = PwCalculation.process()


class EquationOfStates(FragmentedWorkfunction):
    @classmethod
    def _define(cls, spec):
        spec.input("structure", valid_type=StructureData)
        spec.input("start", valid_type=NumericType, default=Float(0.96))
        spec.input("delta", valid_type=NumericType, default=Float(0.02))
        spec.input("end", valid_type=NumericType, default=Float(1.04))
        spec.input("code", valid_type=SimpleData)
        spec.input("pseudo_family", valid_type=SimpleData)
        spec.outline(
            cls.run_pw,
            cls.plot_eos
        )

    def run_pw(self, ctx):
        calcs = {}

        scale = self.inputs.start
        while scale <= self.inputs.end:
            scaled = rescale(self.inputs.structure, scale)

            inputs = generate_scf_input_params(
                scaled, self.inputs.code, self.inputs.pseudo_family)

            # Launch the code
            future = self.submit(PwProcess, inputs)
            #print scale.value, future.pid
            # Store the future
            calcs["s_{}".format(scale)] = future
            scale = scale + self.inputs.delta

        # Ask the workflow to continue when the results are ready and store them
        # in the context
        return ResultToContext(**calcs)

    def plot_eos(self, ctx):
        for label in ctx:
            if "s_" in label:
                print "{} {}".format(
                    label, ctx[label]['output_parameters'].dict.energy)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description='Equation of states example.')
    parser.add_argument('--structure', type=int, dest='structure_pk',
                        help='The structure pk to use.', required=True)
    parser.add_argument('--range', type=str, dest='range',
                        help='The scale range of the equation of states '
                             'calculation written as [start]:[end]:[delta]',
                             default='0.96:1.04:0.02')
    parser.add_argument('--pseudo', type=str, dest='pseudo',
                        help='The pseudopotential family', required=True)
    parser.add_argument('--code', type=str, dest='code',
                        help='The codename to use', required=True)

    args = parser.parse_args()
    start, end, delta = args.range.split(':')

    # Get the structure from the calculation
    # run(EquationOfStates, structure=load_node(args.structure_pk),
    #     start=Float(start), end=Float(end), delta=Float(delta),
    #     code=Str(args.code), pseudo_family=Str(args.pseudo))
    run(EquationOfStates, structure=load_node(args.structure_pk),
        start=Float(start), end=Float(end), delta=Float(delta),
        code=Str(args.code), pseudo_family=Str(args.pseudo))





# -*- coding: utf-8 -*-
from aiida.backends.utils import load_dbenv, is_dbenv_loaded

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__authors__ = "The AiiDA team."
__version__ = "0.7.0.1"

if not is_dbenv_loaded():
    load_dbenv()

import ase
from aiida.workflows2.run import async
from aiida.orm import DataFactory
from aiida.workflows2.db_types import Float, Str
from aiida.workflows2.wf import wf
from aiida.workflows2.run import async
from examples.workflows2.common import run_scf
from aiida.workflows2.defaults import registry

@wf
def create_diamond_fcc(element,alat):
    """
    Workfunction to create a diamond crystal structure with a given element.

    :param element: The element to create the structure with.
    :return: The structure.
    """
    from numpy import array
    the_cell = array([[0., 0.5, 0.5],
                [0.5, 0., 0.5],
                [0.5, 0.5, 0.]]) * alat
    StructureData = DataFactory("structure")
    structure = StructureData(cell=the_cell)
    structure.append_atom(position=(0., 0., 0.), symbols=str(element))
    structure.append_atom(position=(0.25 * alat, 0.25 * alat, 0.25 * alat), symbols=str(element))
    return structure


@wf
def rescale(structure, scale):
    """
    Workfunction to rescale a structure

    :param structure: An AiiDA structure to rescale
    :param scale: The scale factor
    :return: The rescaled structure
    """
    the_ase = structure.get_ase()
    new_ase = the_ase.copy()
    new_ase.set_cell(the_ase.get_cell() * float(scale), scale_atoms=True)
    new_structure = DataFactory('structure')(ase=new_ase)
    return new_structure


@wf
def calc_energies(codename, pseudo_family):
    print("Calculating energies, my pk is '{}'".format(registry.current_pid))

    futures = {}
    for element, scale in [("Si", 5.41)]:
        structure = create_diamond_fcc(Str(element), Float(1.))
        structure = rescale(structure, Float(scale))
        print("Running {} scf calculation.".format(element))
        futures[element] = async(run_scf, structure, codename, pseudo_family)

    print("Waiting for calculations to finish.")
    outs = {element: Float(future.result()['output_parameters'].dict.energy)
            for element, future in futures.iteritems()}
    print("Calculations finished.")
    return outs


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description='Energy calculation example.')
    parser.add_argument('--pseudo', type=str, dest='pseudo', required=True,
                        help='The pseudopotential family')
    parser.add_argument('--code', type=str, dest='code', required=True,
                        help='The codename to use')

    args = parser.parse_args()
    print(calc_energies(Str(args.code), Str(args.pseudo)))



# -*- coding: utf-8 -*-
"""
Command line commands for the main executable 'verdi' of aiida

If you want to define a new command line parameter, just define a new
class inheriting from VerdiCommand, and define a run(self,*args) method
accepting a variable-length number of parameters args
(the command-line parameters), which will be invoked when
this executable is called as
verdi NAME

Don't forget to add the docstring to the class: the first line will be the
short description, the following ones the long description.
"""
import sys
import os
import contextlib

import aiida
from aiida.common.exceptions import (
    AiidaException, ConfigurationError, ProfileConfigurationError)
from aiida.cmdline.baseclass import VerdiCommand, VerdiCommandRouter
from aiida.cmdline import pass_to_django_manage
from aiida.backends import settings as settings_profile

# Import here from other files; once imported, it will be found and
# used as a command-line parameter
from aiida.cmdline.commands.user import User
from aiida.cmdline.commands.calculation import Calculation
from aiida.cmdline.commands.code import Code
from aiida.cmdline.commands.computer import Computer
from aiida.cmdline.commands.daemon import Daemon
from aiida.cmdline.commands.data import Data
from aiida.cmdline.commands.devel import Devel
from aiida.cmdline.commands.exportfile import Export
from aiida.cmdline.commands.group import Group
from aiida.cmdline.commands.graph import Graph
from aiida.cmdline.commands.importfile import Import
from aiida.cmdline.commands.node import Node
from aiida.cmdline.commands.profile import Profile
from aiida.cmdline.commands.workflow import Workflow
from aiida.cmdline.commands.workflow2 import Workflow2
from aiida.cmdline.commands.comment import Comment
from aiida.cmdline.commands.shell import Shell
from aiida.cmdline import execname

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


class ProfileParsingException(AiidaException):
    """
    Exception raised when parsing the profile command line option, if only
    -p is provided, and no profile is specified
    """

    def __init__(self, *args, **kwargs):
        self.minus_p_provided = kwargs.pop('minus_p_provided', False)

        super(ProfileParsingException, self).__init__(*args, **kwargs)


def parse_profile(argv, merge_equal=False):
    """
    Parse the argv to see if a profile has been specified, return it with the
    command position shift (index where the commands start)

    :param merge_equal: if True, merge things like
      ('verdi', '--profile', '=', 'x', 'y') to ('verdi', '--profile=x', 'y')
      but then return the correct index for the original array.

    :raise ProfileParsingException: if there is only 'verdi' specified, or
      if only 'verdi -p' (in these cases, one has respectively
      exception.minus_p_provided equal to False or True)
    """
    if merge_equal:
        if len(argv) >= 3:
            if argv[1] == '--profile' and argv[2] == '=':
                internal_argv = [argv[0], "".join(argv[1:4])] + list(argv[4:])
                shift = 2
            else:
                internal_argv = list(argv)
                shift = 0
        else:
            internal_argv = list(argv)
            shift = 0
    else:
        internal_argv = list(argv)
        shift = 0

    profile = None  # Use default profile if nothing is specified
    command_position = 1  # If there is no profile option
    try:
        profile_switch = internal_argv[1]
    except IndexError:
        raise ProfileParsingException(minus_p_provided=False)
    long_profile_prefix = '--profile='
    if profile_switch == '-p':
        try:
            profile = internal_argv[2]
        except IndexError:
            raise ProfileParsingException(minus_p_provided=True)
        command_position = 3
    elif profile_switch.startswith(long_profile_prefix):
        profile = profile_switch[len(long_profile_prefix):]
        command_position = 2
    else:
        # No profile switch, continue using argv[1] as the command name
        pass

    return profile, command_position + shift


@contextlib.contextmanager
def update_environment(new_argv):
    """
    Used as a context manager, changes sys.argv with the
    new_argv argument, and restores it upon exit.
    """
    import sys

    _argv = sys.argv[:]
    sys.argv = new_argv[:]
    yield

    # Restore old parameters when exiting from the context manager
    sys.argv = _argv


########################################################################
# HERE STARTS THE COMMAND FUNCTION LIST
########################################################################

class CompletionCommand(VerdiCommand):
    """
    Return the bash completion function to put in ~/.bashrc

    This command prints on screen the function to be inserted in
    your .bashrc command. You can copy and paste the output, or simply
    add
    eval "`verdi completioncommand`"
    to your .bashrc, *AFTER* having added the aiida/bin directory to the path.
    """

    def run(self, *args):
        """
        I put the documentation here, and I don't print it, so we
        don't clutter too much the .bashrc.

        * "${THE_WORDS[@]}" (with the @) puts each element as a different
          parameter; note that the variable expansion etc. is performed

        * I add a 'x' at the end and then remove it; in this way, $( ) will
          not remove trailing spaces

        * If the completion command did not print anything, we use
          the default bash completion for filenames

        * If instead the code prints something empty, thanks to the workaround
          above $OUTPUT is not empty, so we do go the the 'else' case
          and then, no substitution is suggested.
        """

        print r"""
function _aiida_verdi_completion
{
    OUTPUT=$( $1 completion "$COMP_CWORD" "${COMP_WORDS[@]}" ; echo 'x')
    OUTPUT=${OUTPUT%x}
    if [ -z "$OUTPUT" ]
    then
    # Only newline is a valid separator
        local IFS=$'\n'

        COMPREPLY=( $(compgen -o default -- "${COMP_WORDS[COMP_CWORD]}" ) )
    # Add either a slash or a space, depending on whether it is a folder
    # or a file. printf %q escapes the filename if there are spaces.
        for ((i=0; i < ${#COMPREPLY[@]}; i++)); do
            [ -d "${COMPREPLY[$i]}" ] && \
               COMPREPLY[$i]=$(printf %q%s "${COMPREPLY[$i]}" "/") || \
               COMPREPLY[$i]=$(printf %q%s "${COMPREPLY[$i]}" " ")
        done

    else
        COMPREPLY=( $(compgen -W "$OUTPUT" -- "${COMP_WORDS[COMP_CWORD]}" ) )
        # Always add a space after each command
        for ((i=0; i < ${#COMPREPLY[@]}; i++)); do
            COMPREPLY[$i]="${COMPREPLY[$i]} "
        done
    fi
}
complete -o nospace -F _aiida_verdi_completion verdi
"""

    def complete(self, subargs_idx, subargs):
        # disable further completion
        print ""


class Completion(VerdiCommand):
    """
    Manage bash completion

    Return a list of available commands, separated by spaces.
    Calls the correct function of the command if the TAB has been
    pressed after the first command.

    Returning without printing will use the default bash completion.
    """

    # TODO: manage completion at a deeper level

    def run(self, *args):
        try:
            cword = int(args[0])
            if cword <= 0:
                cword = 1
        except IndexError:
            cword = 1
        except ValueError:
            return

        try:
            profile, command_position = parse_profile(args[1:],
                                                      merge_equal=True)
        except ProfileParsingException as e:
            cword_offset = 0
        else:
            cword_offset = command_position - 1

        if cword == 1 + cword_offset:
            print " ".join(sorted(short_doc.keys()))
            return
        else:
            try:
                # args[0] is cword;
                # args[1] is the executable (verdi)
                # args[2] is the command for verdi
                # args[3:] are the following subargs
                command = args[2 + cword_offset]
            except IndexError:
                return
            try:
                CommandClass = list_commands[command]
            except KeyError:
                return
            CommandClass().complete(subargs_idx=cword - 2 - cword_offset,
                                    subargs=args[3 + cword_offset:])


class ListParams(VerdiCommand):
    """
    List available commands

    List available commands and their short description.
    For the long description, use the 'help' command.
    """

    def run(self, *args):
        print get_listparams()


class Help(VerdiCommand):
    """
    Describe a specific command

    Pass a further argument to get a description of a given command.
    """

    def run(self, *args):
        try:
            command = args[0]
        except IndexError:
            print get_listparams()
            print ""
            print (
            "Before each command you can specify the AiiDA profile to use,"
            " with 'verdi -p <profile> <command>' or "
            "'verdi --profile=<profile> <command>'")
            print ""
            print ("Use '{} help <command>' for more information "
                   "on a specific command.".format(execname))
            sys.exit(1)

        if command in short_doc:
            print "Description for '%s %s'" % (execname, command)
            print ""
            print "**", short_doc[command]
            if command in long_doc:
                print long_doc[command]
        else:
            print >> sys.stderr, (
                "{}: '{}' is not a valid command. "
                "See '{} help' for more help.".format(
                    execname, command, execname))
            get_command_suggestion(command)
            sys.exit(1)

    def complete(self, subargs_idx, subargs):
        if subargs_idx == 0:
            print " ".join(sorted(short_doc.keys()))
        else:
            print ""


class Install(VerdiCommand):
    """
    Install/setup aiida for the current user

    This command creates the ~/.aiida folder in the home directory
    of the user, interactively asks for the database settings and
    the repository location, does a setup of the daemon and runs
    a migrate command to create/setup the database.
    """

    def run(self, *args):
        from aiida.common.setup import (create_base_dirs, create_configuration,
                                        set_default_profile, DEFAULT_UMASK)
        from aiida.backends.profile import BACKEND_SQLA, BACKEND_DJANGO
        from aiida.backends.utils import set_backend_type, get_backend_type
        from aiida.common.exceptions import InvalidOperation

        cmdline_args = list(args)

        only_user_config = False
        try:
            cmdline_args.remove('--only-config')
            only_user_config = True
        except ValueError:
            # Parameter not provided
            pass

        if cmdline_args:
            print >> sys.stderr, "Unknown parameters on the command line: "
            print >> sys.stderr, ", ".join(cmdline_args)
            sys.exit(1)

        # create the directories to store the configuration files
        create_base_dirs()
        # gprofile = 'default' if profile is None else profile
        gprofile = 'default' if settings_profile.AIIDADB_PROFILE is None \
            else settings_profile.AIIDADB_PROFILE

        created_conf = None
        # ask and store the configuration of the DB
        try:
            created_conf = create_configuration(profile=gprofile)
        except ValueError as e:
            print >> sys.stderr, "Error during configuration: {}".format(
                e.message)
            sys.exit(1)

        # set default DB profiles
        set_default_profile('verdi', gprofile, force_rewrite=False)
        set_default_profile('daemon', gprofile, force_rewrite=False)

        if only_user_config:
            print ("Only user configuration requested, "
                   "skipping the migrate command")
        else:
            print "Executing now a migrate command..."

            backend_choice = created_conf['AIIDADB_BACKEND']
            if backend_choice == BACKEND_DJANGO:
                print("...for Django backend")
                # The correct profile is selected within load_dbenv.
                # Setting os.umask here since sqlite database gets created in
                # this step.
                old_umask = os.umask(DEFAULT_UMASK)

                # This check should be done more properly
                # try:
                #     backend_type = get_backend_type()
                # except KeyError:
                #     backend_type = None
                #
                # if backend_type is not None and backend_type != BACKEND_DJANGO:
                #     raise InvalidOperation("An already existing database found"
                #                            "and a different than the selected"
                #                            "backend was used for its "
                #                            "management.")

                try:
                    pass_to_django_manage([execname, 'migrate'],
                                          profile=gprofile)
                finally:
                    os.umask(old_umask)

                set_backend_type(BACKEND_DJANGO)

            elif backend_choice == BACKEND_SQLA:
                print("...for SQLAlchemy backend")
                from aiida.backends.sqlalchemy.models.base import Base
                from aiida.backends.sqlalchemy.utils import (get_engine,
                                                             install_tc)
                from aiida.common.setup import get_profile_config
                from aiida import is_dbenv_loaded, load_dbenv

                if not is_dbenv_loaded():
                    load_dbenv()

                # This check should be done more properly
                # try:
                #     backend_type = get_backend_type()
                # except KeyError:
                #     backend_type = None
                #
                # if backend_type is not None and backend_type != BACKEND_SQLA:
                #     raise InvalidOperation("An already existing database found"
                #                            "and a different than the selected"
                #                            "backend was used for its "
                #                            "management.")

                # Those import are necessary for SQLAlchemy to correctly create
                # the needed database tables.
                from aiida.backends.sqlalchemy.models.authinfo import (
                    DbAuthInfo)
                from aiida.backends.sqlalchemy.models.comment import DbComment
                from aiida.backends.sqlalchemy.models.computer import (
                    DbComputer)
                from aiida.backends.sqlalchemy.models.group import (
                    DbGroup, table_groups_nodes)
                from aiida.backends.sqlalchemy.models.lock import DbLock
                from aiida.backends.sqlalchemy.models.log import DbLog
                from aiida.backends.sqlalchemy.models.node import (
                    DbLink, DbNode, DbPath, DbCalcState)
                from aiida.backends.sqlalchemy.models.user import DbUser
                from aiida.backends.sqlalchemy.models.workflow import (
                    DbWorkflow, DbWorkflowData, DbWorkflowStep)
                from aiida.backends.sqlalchemy.models.settings import DbSetting

                connection = get_engine(get_profile_config(gprofile))
                Base.metadata.create_all(connection)
                install_tc(connection)

                set_backend_type(BACKEND_SQLA)

            else:
                raise InvalidOperation("Not supported backend selected.")

        print "Database was created successfully"

        # I create here the default user
        print "Loading new environment..."
        if only_user_config:
            from aiida.backends.utils import load_dbenv, is_dbenv_loaded
            # db environment has not been loaded in this case
            if not is_dbenv_loaded():
                load_dbenv()

        from aiida.common.setup import DEFAULT_AIIDA_USER
        from aiida.orm.user import User as AiiDAUser

        if not AiiDAUser.search_for_users(email=DEFAULT_AIIDA_USER):
            print "Installing default AiiDA user..."
            nuser = AiiDAUser(email=DEFAULT_AIIDA_USER)
            nuser.first_name = "AiiDA"
            nuser.last_name = "Daemon"
            nuser.is_staff = True
            nuser.is_active = True
            nuser.is_superuser = True
            nuser.force_save()

        from aiida.common.utils import get_configured_user_email
        email = get_configured_user_email()
        print "Starting user configuration for {}...".format(email)
        if email == DEFAULT_AIIDA_USER:
            print "You set up AiiDA using the default Daemon email ({}),".format(
                email)
            print "therefore no further user configuration will be asked."
        else:
            # Ask to configure the new user
            User().user_configure(email)

        print "Install finished."

    def complete(self, subargs_idx, subargs):
        """
        No completion after 'verdi install'.
        """
        print ""


class Runserver(VerdiCommand):
    """
    Run the AiiDA webserver on localhost

    This command runs the webserver on the default port. Further command line
    options are passed to the Django manage runserver command
    """

    def run(self, *args):
        pass_to_django_manage([execname, 'runserver'] + list(args))


class Run(VerdiCommand):
    """
    Execute an AiiDA script
    """

    def run(self, *args):
        from aiida.backends.utils import load_dbenv,is_dbenv_loaded

        if not is_dbenv_loaded():
            load_dbenv()
        import argparse
        from aiida.cmdline.commands.shell import default_modules_list
        import aiida.orm.autogroup
        from aiida.orm.autogroup import Autogroup

        parser = argparse.ArgumentParser(
            prog=self.get_full_command_name(),
            description='Execute an AiiDA script.')
        parser.add_argument('-g', '--group', type=bool, default=True,
                            help='Enables the autogrouping, default = True')
        parser.add_argument('-n', '--groupname', type=str, default=None,
                            help='Specify the name of the auto group')
        # parser.add_argument('-o','--grouponly', type=str, nargs='+', default=['all'],
        #                            help='Limit the grouping to specific classes (by default, all classes are grouped')
        parser.add_argument('-e', '--exclude', type=str, nargs='+', default=[],
                            help=('Autogroup only specific calculation classes.'
                                  " Select them by their module name.")
                            )
        parser.add_argument('-E', '--excludesubclasses', type=str, nargs='+',
                            default=[],
                            help=('Autogroup only specific calculation classes.'
                                  " Select them by their module name.")
                            )
        parser.add_argument('-i', '--include', type=str, nargs='+',
                            default=['all'],
                            help=('Autogroup only specific data classes.'
                                  " Select them by their module name.")
                            )
        parser.add_argument('-I', '--includesubclasses', type=str, nargs='+',
                            default=[],
                            help=('Autogroup only specific code classes.'
                                  " Select them by their module name.")
                            )
        parser.add_argument('scriptname', metavar='ScriptName', type=str,
                            help='The name of the script you want to execute')
        parser.add_argument('new_args', metavar='ARGS',
                            nargs=argparse.REMAINDER, type=str,
                            help='Further parameters to pass to the script')
        parsed_args = parser.parse_args(args)

        # Prepare the environment for the script to be run
        globals_dict = {
            '__builtins__': globals()['__builtins__'],
            '__name__': '__main__',
            '__file__': parsed_args.scriptname,
            '__doc__': None,
            '__package__': None}

        ## dynamically load modules (the same of verdi shell) - but in
        ## globals_dict, not in the current environment
        for app_mod, model_name, alias in default_modules_list:
            globals_dict["{}".format(alias)] = getattr(
                __import__(app_mod, {}, {}, model_name), model_name)

        if parsed_args.group:
            automatic_group_name = parsed_args.groupname
            if automatic_group_name is None:
                import datetime

                now = datetime.datetime.now()
                automatic_group_name = "Verdi autogroup on " + now.strftime(
                    "%Y-%m-%d %H:%M:%S")

            aiida_verdilib_autogroup = Autogroup()
            aiida_verdilib_autogroup.set_exclude(parsed_args.exclude)
            aiida_verdilib_autogroup.set_include(parsed_args.include)
            aiida_verdilib_autogroup.set_exclude_with_subclasses(
                parsed_args.excludesubclasses)
            aiida_verdilib_autogroup.set_include_with_subclasses(
                parsed_args.includesubclasses)
            aiida_verdilib_autogroup.set_group_name(automatic_group_name)
            ## Note: this is also set in the exec environment!
            ## This is the intended behavior
            aiida.orm.autogroup.current_autogroup = aiida_verdilib_autogroup

        try:
            f = open(parsed_args.scriptname)
        except IOError:
            print >> sys.stderr, "{}: Unable to load file '{}'".format(
                self.get_full_command_name(), parsed_args.scriptname)
            sys.exit(1)
        else:
            try:
                # Must add also argv[0]
                new_argv = [parsed_args.scriptname] + parsed_args.new_args
                with update_environment(new_argv=new_argv):
                    # Add local folder to sys.path
                    sys.path.insert(0, os.path.abspath(os.curdir))
                    # Pass only globals_dict
                    exec (f, globals_dict)
                    # print sys.argv
            except SystemExit as e:
                ## Script called sys.exit()
                # print sys.argv, "(sys.exit {})".format(e.message)

                ## Note: remember to re-raise, the exception to have
                ## the error code properly returned at the end!
                raise
            finally:
                f.close()


########################################################################
# HERE ENDS THE COMMAND FUNCTION LIST
########################################################################
# From here on: utility functions

def get_listparams():
    """
    Return a string with the list of parameters, to be printed

    The advantage of this function is that the calling routine can
    choose to print it on stdout or stderr, depending on the needs.
    """
    max_length = max(len(i) for i in short_doc.keys())

    name_desc = [(cmd.ljust(max_length + 2), desc.strip())
                 for cmd, desc in short_doc.iteritems()]

    name_desc = sorted(name_desc)

    return ("List of the most relevant available commands:" + os.linesep +
            os.linesep.join(["  * {} {}".format(name, desc)
                             for name, desc in name_desc]))


def get_command_suggestion(command):
    """
    A function that prints on stderr a list of similar commands
    """
    import difflib

    similar_cmds = difflib.get_close_matches(command, short_doc.keys())
    if similar_cmds:
        print >> sys.stderr, ""
        print >> sys.stderr, "Did you mean this?"
        print >> sys.stderr, "\n".join(["     {}".format(i)
                                        for i in similar_cmds])


def print_usage(execname):
    print >> sys.stderr, ("Usage: {} [--profile=PROFILENAME|-p PROFILENAME] "
                          "COMMAND [<args>]".format(execname))
    print >> sys.stderr, ""
    print >> sys.stderr, get_listparams()
    print >> sys.stderr, "See '{} help' for more help.".format(execname)


def exec_from_cmdline(argv):
    """
    The main function to be called. Pass as parameter the sys.argv.
    """
    ### This piece of code takes care of creating a list of valid
    ### commands and of their docstrings for dynamic management of
    ### the code.
    ### It defines a few global variables

    global execname
    global list_commands
    global short_doc
    global long_doc

    # import itself
    from aiida.cmdline import verdilib
    import inspect

    # List of command names that should be hidden or not completed.
    hidden_commands = ['completion', 'completioncommand', 'listparams']

    # Retrieve the list of commands
    verdilib_namespace = verdilib.__dict__

    list_commands = {v.get_command_name(): v for v in
                     verdilib_namespace.itervalues()
                     if inspect.isclass(v) and not v == VerdiCommand and
                     issubclass(v, VerdiCommand)
                     and not v.__name__.startswith('_')
                     and not v._abstract}

    # Retrieve the list of docstrings, managing correctly the
    # case of empty docstrings. Each value is a list of lines
    raw_docstrings = {k: (v.__doc__ if v.__doc__ else "").splitlines()
                      for k, v in list_commands.iteritems()}

    short_doc = {}
    long_doc = {}
    for k, v in raw_docstrings.iteritems():
        if k in hidden_commands:
            continue
        lines = [l.strip() for l in v]
        empty_lines = [bool(l) for l in lines]
        try:
            first_idx = empty_lines.index(True)  # The first non-empty line
        except ValueError:
            # All False
            short_doc[k] = "No description available"
            long_doc[k] = ""
            continue
        short_doc[k] = lines[first_idx]
        long_doc[k] = os.linesep.join(lines[first_idx + 1:])

    execname = os.path.basename(argv[0])

    try:
        profile, command_position = parse_profile(argv)
    except ProfileParsingException as e:
        print_usage(execname)
        sys.exit(1)

    # We now set the internal variable, if needed
    if profile is not None:
        settings_profile.AIIDADB_PROFILE = profile
    # I set the process to verdi
    settings_profile.CURRENT_AIIDADB_PROCESS = "verdi"

    # Finally, we parse the commands and their options
    try:
        command = argv[command_position]
    except IndexError:
        print_usage(execname)
        sys.exit(1)

    try:
        if command in list_commands:
            CommandClass = list_commands[command]()
            CommandClass.run(*argv[command_position + 1:])
        else:
            print >> sys.stderr, ("{}: '{}' is not a valid command. "
                                  "See '{} help' for more help.".format(
                execname, command, execname))
            get_command_suggestion(command)
            sys.exit(1)
    except ProfileConfigurationError as e:
        print >> sys.stderr, "The profile specified is not valid!"
        print >> sys.stderr, e.message
        sys.exit(1)



# -*- coding: utf-8 -*-

from datetime import datetime

from querybuilder_base import (
    AbstractQueryBuilder,
    datetime,
    InputValidationError
)

from aiida.backends.querybuild.dummy_model import (
    # Tables:
    DbNode      as DummyNode,
    DbLink      as DummyLink,
    DbCalcState as DummyState,
    DbPath      as DummyPath,
    DbUser      as DummyUser,
    DbComputer  as DummyComputer,
    DbGroup     as DummyGroup,
    DbExtra     as DummyExtra,
    DbAttribute as DummyAttribute,
    table_groups_nodes  as Dummy_table_groups_nodes,
    session,                             # session with DB
)

from aiida.backends.djsite.db.models import DbAttribute, DbExtra, ObjectDoesNotExist


from aiida.backends.querybuild.sa_init import (
    and_, or_, aliased,      # Queryfuncs
    cast, Float, Integer, Boolean, DateTime,
    case, exists, join, select, exists
)

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__authors__ = "The AiiDA team."
__version__ = "0.7.0.1"

class QueryBuilder(AbstractQueryBuilder):

    def __init__(self, *args, **kwargs):
        from aiida.orm.implementation.django.node import Node as AiidaNode
        from aiida.orm.implementation.django.group import Group as AiidaGroup
        from aiida.orm.implementation.django.computer import Computer as AiidaComputer
        from aiida.orm.implementation.django.user import User as AiidaUser

        self.Link               = DummyLink
        self.Path               = DummyPath
        self.Node               = DummyNode
        self.Computer           = DummyComputer
        self.User               = DummyUser
        self.Group              = DummyGroup
        self.table_groups_nodes = Dummy_table_groups_nodes
        self.AiidaNode          = AiidaNode
        self.AiidaGroup         = AiidaGroup
        self.AiidaComputer      = AiidaComputer
        self.AiidaUser          = AiidaUser

        super(QueryBuilder, self).__init__(*args, **kwargs)


    def _get_aiida_res(self, key, res):
        """
        Some instance returned by ORM (django or SA) need to be converted
        to Aiida instances (eg nodes)

        :param res: the result returned by the query
        :param key: the key that this entry would be return with

        :returns: an aiida-compatible instance
        """
        if key.startswith('attributes'):
            try:
                returnval = DbAttribute.objects.get(id=res).getvalue()
            except ObjectDoesNotExist:
                returnval = None
        elif key.startswith('extras'):
            try:
                returnval = DbExtra.objects.get(id=res).getvalue()
            except ObjectDoesNotExist:
                returnval = None
        elif isinstance(res, (self.Group, self.Node, self.Computer, self.User)):
            returnval =  res.get_aiida_class()
        else:
            returnval = res
        return returnval


    @staticmethod
    def _get_session():
        return session

    @classmethod
    def _get_filter_expr_from_attributes(
            cls, operator, value, attr_key,
            column=None, column_name=None,
            alias=None):

        def get_attribute_db_column(mapped_class, dtype, castas=None):
            if dtype == 't':
                mapped_entity = mapped_class.tval
            elif dtype == 'b':
                mapped_entity = mapped_class.bval
                #~ mapped_entity = cast(mapped_class.value_str, Boolean)
            elif dtype == 'f':
                mapped_entity = mapped_class.fval
                #~ mapped_entity = cast(mapped_class.value_str, Float)
            elif dtype == 'i':
                mapped_entity = mapped_class.ival
                #~ mapped_entity = cast(mapped_class.value_str, Integer)
            elif dtype == 'd':
                mapped_entity = mapped_class.dval
            else:
                raise InputValidationError(
                    "I don't know what to do with dtype {}".format(dtype)
                )
            if castas == 't':
                mapped_entity = cast(mapped_entity, String)
            elif castas == 'f':
                mapped_entity = cast(mapped_entity, Float)

            return mapped_entity
        if column:
            mapped_class = db_column.prop.mapper.class_
        else:
            column = getattr(alias, column_name)
            mapped_class = column.prop.mapper.class_
        # Ok, so we have an attribute key here.
        # Unless cast is specified, will try to infer my self where the value
        # is stored
        # Datetime -> dval
        # bool -> bval
        # string -> tval
        # integer -> ival, fval (cast ival to float)
        # float -> ival, fval (cast ival to float)
        # If the user specified of_type ??
        # That is basically a query for where the value is sitting
        #   (which db_column in the dbattribtues)
        # If the user specified in what to cast, he wants an operation to
        #   be performed to cast the value to a different type
        if  isinstance(value, (list, tuple)):
            value_type_set = set([type(i) for i in value])
            if len(value_type_set) > 1:
                raise InputValidationError( '{}  contains more than one type'.format(value))
            elif len(value_type_set) == 0:
                raise InputValidationError('Given list is empty, cannot determine type')
            else:
                value_to_consider = value[0]
        else:
            value_to_consider = value


        # First cases, I maybe need not do anything but just count the
        # number of entries
        if operator in ('of_length', 'shorter', 'longer'):
            raise NotImplementedError(
                "Filtering by lengths of arrays or lists is not implemented\n"
                "in the Django-Backend"
            )
        elif operator  == 'of_type':
            raise NotImplementedError(
                "Filtering by type is not implemented\n"
                "in the Django-Backend"
            )
        elif operator == 'contains':
            raise NotImplementedError(
                "Contains is not implemented in the Django-backend"
            )


        elif operator=='has_key':
            if issubclass(mapped_class, DummyAttribute):
                expr = alias.attributes.any(mapped_class.key == '.'.join(attr_key+[value]))
            elif issubclass(mapped_class, DummyExtra):
                expr = alias.extras.any(mapped_class.key == '.'.join(attr_key+[value]))
            else:
                raise Exception("I was given {} as an attribute base class".format(mapped_class))

        else:
            types_n_casts = []
            if isinstance(value_to_consider, basestring):
                types_n_casts.append(('t', None))
            elif isinstance(value_to_consider, bool):
                types_n_casts.append(('b', None))
            elif isinstance(value_to_consider, (int, float)):
                types_n_casts.append(('f', None))
                types_n_casts.append(('i', 'f'))
            elif isinstance(value_to_consider, datetime.datetime):
                types_n_casts.append(('d', None))

            expressions = []
            for dtype, castas in types_n_casts:
                try:
                    expressions.append(
                        cls._get_filter_expr(
                            operator, value, attr_key=[],
                            column=get_attribute_db_column(mapped_class, dtype, castas=castas),
                            is_attribute=False
                        )
                    )
                except InputValidationError as e:
                    raise e

            actual_attr_key = '.'.join(attr_key)
            expr = column.any(and_(
                    mapped_class.key == actual_attr_key,
                    or_(*expressions)
                )
            )
        return expr



    def _get_projectable_attribute(
            self, alias, column_name, attrpath,
            cast=None, **kwargs
        ):
        if cast is not None:
            raise NotImplementedError(
                "Casting is not implemented in the Django backend"
            )
        if not attrpath:
            raise NotImplementedError(
                "Cannot project all attributes in the Django backend\n"
                "(You did not provide a key)"
            )

        aliased_attributes = aliased(getattr(alias, column_name).prop.mapper.class_)

        if not issubclass(alias._aliased_insp.class_,self.Node):
            NotImplementedError(
                "Other classes than Nodes are not implemented yet"
            )

        attrkey = '.'.join(attrpath)


        exists_stmt = exists(select([1], correlate=True).select_from(
                aliased_attributes
            ).where(and_(
                aliased_attributes.key==attrkey,
                aliased_attributes.dbnode_id==alias.id
            )))

        select_stmt = select(
                [aliased_attributes.id], correlate=True
            ).select_from(aliased_attributes).where(and_(
                aliased_attributes.key==attrkey,
                aliased_attributes.dbnode_id==alias.id
            )).label('miao')

        entity = case([(exists_stmt, select_stmt), ], else_=None)

        return entity



# -*- coding: utf-8 -*-
"""
This module defines the classes related to band structures or dispersions
in a Brillouin zone, and how to operate on them.
"""

from aiida.orm.data.array.kpoints import KpointsData
import numpy
from string import Template
from aiida.common.exceptions import ValidationError

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


# TODO: set and get bands could have more functionalities: how do I know the number of bands for example?

def find_bandgap(bandsdata, number_electrons=None, fermi_energy=None):
    """
    Tries to guess whether the bandsdata represent an insulator.
    This method is meant to be used only for electronic bands (not phonons)
    By default, it will try to use the occupations to guess the number of
    electrons and find the Fermi Energy, otherwise, it can be provided
    explicitely.
    Also, there is an implicit assumption that the kpoints grid is
    "sufficiently" dense, so that the bandsdata are not missing the
    intersection between valence and conduction band if present.
    Use this function with care!

    :param (float) number_electrons: (optional) number of electrons in the unit cell
    :param (float) fermi_energy: (optional) value of the fermi energy.

    :note: By default, the algorithm uses the occupations array
      to guess the number of electrons and the occupied bands. This is to be
      used with care, because the occupations could be smeared so at a
      non-zero temperature, with the unwanted effect that the conduction bands
      might be occupied in an insulator.
      Prefer to pass the number_of_electrons explicitly

    :note: Only one between number_electrons and fermi_energy can be specified at the
      same time.

    :return: (is_insulator, gap), where is_insulator is a boolean, and gap a
             float. The gap is None in case of a metal, zero when the homo is
             equal to the lumo (e.g. in semi-metals).
    """

    def nint(num):
        """
        Stable rounding function
        """
        if (num > 0):
            return int(num + .5)
        else:
            return int(num - .5)

    if fermi_energy and number_electrons:
        raise ValueError("Specify either the number of electrons or the "
                         "Fermi energy, but not both")

    try:
        stored_bands = bandsdata.get_bands()
    except KeyError:
        raise KeyError("Cannot do much of a band analysis without bands")

    if len(stored_bands.shape) == 3:
        # I write the algorithm for the generic case of having both the
        # spin up and spin down array

        # put all spins on one band per kpoint
        bands = numpy.concatenate([_ for _ in stored_bands], axis=1)
    else:
        bands = stored_bands

    # analysis on occupations:
    if fermi_energy is None:

        num_kpoints = len(bands)

        if number_electrons is None:
            try:
                _, stored_occupations = bandsdata.get_bands(also_occupations=True)
            except KeyError:
                raise KeyError("Cannot determine metallicity if I don't have "
                               "either fermi energy, or occupations")

            # put the occupations in the same order of bands, also in case of multiple bands
            if len(stored_occupations.shape) == 3:
                # I write the algorithm for the generic case of having both the
                # spin up and spin down array

                # put all spins on one band per kpoint
                occupations = numpy.concatenate([_ for _ in stored_occupations], axis=1)
            else:
                occupations = stored_occupations

            # now sort the bands by energy
            # Note: I am sort of assuming that I have an electronic ground state

            # sort the bands by energy, and reorder the occupations accordingly
            # since after joining the two spins, I might have unsorted stuff
            bands, occupations = [numpy.array(y) for y in zip(*[zip(*j) for j in
                                                                [sorted(zip(i[0].tolist(), i[1].tolist()),
                                                                        key=lambda x: x[0])
                                                                 for i in zip(bands, occupations)]])]
            number_electrons = int(round(sum([sum(i) for i in occupations]) / num_kpoints))

            homo_indexes = [numpy.where(numpy.array([nint(_) for _ in x]) > 0)[0][-1] for x in occupations]
            if len(set(homo_indexes)) > 1:  # there must be intersections of valence and conduction bands
                return False, None
            else:
                homo = [_[0][_[1]] for _ in zip(bands, homo_indexes)]
                try:
                    lumo = [_[0][_[1] + 1] for _ in zip(bands, homo_indexes)]
                except IndexError:
                    raise ValueError("To understand if it is a metal or insulator, "
                                     "need more bands than n_band=number_electrons")

        else:
            bands = numpy.sort(bands)
            number_electrons = int(number_electrons)

            # find the zero-temperature occupation per band (1 for spin-polarized
            # calculation, 2 otherwise)
            number_electrons_per_band = 4 - len(stored_bands.shape)  # 1 or 2
            # gather the energies of the homo band, for every kpoint
            homo = [i[number_electrons / number_electrons_per_band - 1] for i in bands]  # take the nth level
            try:
                # gather the energies of the lumo band, for every kpoint
                lumo = [i[number_electrons / number_electrons_per_band] for i in bands]  # take the n+1th level
            except IndexError:
                raise ValueError("To understand if it is a metal or insulator, "
                                 "need more bands than n_band=number_electrons")

        if number_electrons % 2 == 1 and len(stored_bands.shape) == 2:
            # if #electrons is odd and we have a non spin polarized calculation
            # it must be a metal and I don't need further checks
            return False, None

        # if the nth band crosses the (n+1)th, it is an insulator
        gap = min(lumo) - max(homo)
        if gap == 0.:
            return False, 0.
        elif gap < 0.:
            return False, None
        else:
            return True, gap

    # analysis on the fermi energy
    else:
        # reorganize the bands, rather than per kpoint, per energy level

        # I need the bands sorted by energy
        bands.sort()

        levels = bands.transpose()
        max_mins = [(max(i), min(i)) for i in levels]

        if fermi_energy > bands.max():
            raise ValueError("The Fermi energy is above all band energies, "
                             "don't know what to do")
        if fermi_energy < bands.min():
            raise ValueError("The Fermi energy is below all band energies, "
                             "don't know what to do.")

        # one band is crossed by the fermi energy
        if any(i[1] < fermi_energy and fermi_energy < i[0] for i in max_mins):
            return False, None

        # case of semimetals, fermi energy at the crossing of two bands
        # this will only work if the dirac point is computed!
        elif (any(i[0] == fermi_energy for i in max_mins) and
                  any(i[1] == fermi_energy for i in max_mins)):
            return False, 0.
        # insulating case
        else:
            # take the max of the band maxima below the fermi energy
            homo = max([i[0] for i in max_mins if i[0] < fermi_energy])
            # take the min of the band minima above the fermi energy
            lumo = min([i[1] for i in max_mins if i[1] > fermi_energy])
            gap = lumo - homo
            if gap <= 0.:
                raise Exception("Something wrong has been implemented. "
                                "Revise the code!")
            return True, gap


class BandsData(KpointsData):
    """
    Class to handle bands data
    """

    def set_kpointsdata(self, kpointsdata):
        """
        Load the kpoints from a kpoint object.
        :param kpointsdata: an instance of KpointsData class
        """
        if not isinstance(kpointsdata, KpointsData):
            raise ValueError("kpointsdata must be of the KpointsData class")
        try:
            self.cell = kpointsdata.cell
        except AttributeError:
            pass
        try:
            self.pbc = kpointsdata.pbc
        except AttributeError:
            pass
        try:
            self.bravais_lattice = kpointsdata.bravais_lattice
        except AttributeError:
            pass
        try:
            the_kpoints = kpointsdata.get_kpoints()
        except AttributeError:
            the_kpoints = None
        try:
            the_weights = kpointsdata.get_kpoints(also_weights=True)[1]
        except AttributeError:
            the_weights = None
        self.set_kpoints(the_kpoints, weights=the_weights)
        try:
            self.labels = kpointsdata.labels
        except (AttributeError, TypeError):
            self.labels = []

    def _validate_bands_occupations(self, bands, occupations=None, labels=None):
        """
        Validate the list of bands and of occupations before storage.
        Kpoints must be set in advance.
        Bands and occupations must be convertible into arrays of
        Nkpoints x Nbands floats or Nspins x Nkpoints x Nbands; Nkpoints must
        correspond to the number of kpoints.
        """
        try:
            kpoints = self.get_kpoints()
        except AttributeError:
            raise AttributeError("Must first set the kpoints, then the bands")

        the_bands = numpy.array(bands)

        if len(the_bands.shape) not in [2, 3]:
            raise ValueError("Bands must be an array of dimension 2"
                             "([N_kpoints, N_bands]) or of dimension 3 "
                             " ([N_arrays, N_kpoints, N_bands]), found instead {}"
                             .format(len(the_bands.shape)))

        list_of_arrays_to_be_checked = []

        # check that the shape of everything is consistent with the kpoints
        num_kpoints_from_bands = the_bands.shape[0] if len(the_bands.shape) == 2 else the_bands.shape[1]
        if num_kpoints_from_bands != len(kpoints):
            raise ValueError("There must be energy values for every kpoint")

        if occupations is not None:
            the_occupations = numpy.array(occupations)
            if the_occupations.shape != the_bands.shape:
                raise ValueError("Shape of occupations {} different from shape"
                                 "shape of bands {}".format(the_occupations.shape,
                                                            the_bands.shape))

            if not the_bands.dtype.type == numpy.float64:
                list_of_arrays_to_be_checked.append([the_occupations, 'occupations'])
        else:
            the_occupations = None
            # list_of_arrays_to_be_checked = [ [the_bands,'bands'] ]

        # check that there every element is a float
        if not the_bands.dtype.type == numpy.float64:
            list_of_arrays_to_be_checked.append([the_bands, 'bands'])

        for x, msg in list_of_arrays_to_be_checked:
            try:
                [float(_) for _ in x.flatten() if _ is not None]
            except (TypeError, ValueError):
                raise ValueError("The {} array can only contain "
                                 "float or None values".format(msg))

        # check the labels
        if labels is not None:
            if isinstance(labels, basestring):
                the_labels = [str(labels)]
            elif isinstance(labels, (tuple, list)) and all([isinstance(_, basestring) for _ in labels]):
                the_labels = [str(_) for _ in labels]
            else:
                raise ValidationError("Band labels have an unrecognized type ({})"
                                      "but should be a string or a list of strings".format(labels.__class__))

            if len(the_bands.shape) == 2 and len(the_labels) != 1:
                raise ValidationError("More array labels than the number of arrays")
            elif len(the_bands.shape) == 3 and len(the_labels) != the_bands.shape[0]:
                raise ValidationError("More array labels than the number of arrays")
        else:
            the_labels = None

        return the_bands, the_occupations, the_labels

    def set_bands(self, bands, units=None, occupations=None, labels=None):
        """
        Set an array of band energies of dimension (nkpoints x nbands).
        Kpoints must be set in advance. Can contain floats or None.
        :param bands: a list of nkpoints lists of nbands bands, or a 2D array
                of shape (nkpoints x nbands), with band energies for each kpoint
        :param units: optional, energy units
        :param occupations: optional, a 2D list or array of floats of same shape
                as bands, with the occupation associated to each band
        """
        # checks bands and occupations
        the_bands, the_occupations, the_labels = self._validate_bands_occupations(bands,
                                                                                  occupations,
                                                                                  labels)
        # set bands and their units
        self.set_array('bands', the_bands)
        self.units = units

        if the_labels is not None:
            self._set_attr('array_labels', the_labels)

        if the_occupations is not None:
            # set occupations
            self.set_array('occupations', the_occupations)

    @property
    def array_labels(self):
        """
        Get the labels associated with the band arrays
        """
        return self.get_attr('array_labels', None)

    @property
    def units(self):
        """
        Units in which the data in bands were stored. A string
        """
        # return copy.deepcopy(self._pbc)
        return self.get_attr('units')

    @units.setter
    def units(self, value):
        """
        Set the value of pbc, i.e. a tuple of three booleans, indicating if the
        cell is periodic in the 1,2,3 crystal direction
        """
        the_str = str(value)
        self._set_attr('units', the_str)

    def _set_pbc(self, value):
        """
        validate the pbc, then store them
        """
        from aiida.common.exceptions import ModificationNotAllowed
        from aiida.orm.data.structure import get_valid_pbc

        if self.is_stored:
            raise ModificationNotAllowed(
                "The KpointsData object cannot be modified, "
                "it has already been stored")
        the_pbc = get_valid_pbc(value)
        self._set_attr('pbc1', the_pbc[0])
        self._set_attr('pbc2', the_pbc[1])
        self._set_attr('pbc3', the_pbc[2])

    def get_bands(self, also_occupations=False, also_labels=False):
        """
        Returns an array (nkpoints x num_bands or nspins x nkpoints x num_bands)
        of energies.
        :param also_occupations: if True, returns also the occupations array.
            Default = False
        """
        try:
            bands = numpy.array(self.get_array('bands'))
        except KeyError:
            raise AttributeError("No stored bands has been found")

        to_return = [bands]

        if also_occupations:
            try:
                occupations = numpy.array(self.get_array('occupations'))
            except KeyError:
                raise AttributeError('No occupations were set')
            to_return.append(occupations)

        if also_labels:
            to_return.append(self.array_labels)

        if len(to_return) == 1:
            return bands
        else:
            return to_return

    def _exportstring(self, fileformat, path=None, comments=True, cartesian=True,
                      **kwargs):
        """
        Export the bands to a string.
        :param fileformat: format of the file created (e.g. 'agr').
        :param path: absolute path of the file to be created (used for
                additional files, and put in plot_info['filename'])
        :param comments: if True, append some extra informations at the
                beginning of the file.
        :param cartesian: if True, distances (for the x-axis) are computed in
                cartesian coordinates, otherwise they are computed in reciprocal
                coordinates. cartesian=True will fail if no cell has been set.
        :param **kwargs: additional parameters to be passed to the
                _prepare_[fileformat] method

        :note: this function will NOT produce nice plots if:
              - there is no path of kpoints, but a set of isolated points
              - the path is not continuous AND no labels are set
        """
        # TODO: check for None in bands
        import os
        from aiida import get_file_header

        preparer_name = "_prepare_" + fileformat

        # load the x and y's of the graph
        stored_bands = self.get_bands()
        if len(stored_bands.shape) == 2:
            bands = stored_bands
        elif len(stored_bands.shape) == 3:
            bands = numpy.concatenate([_ for _ in stored_bands], axis=1)
        else:
            raise ValueError("Unexpected shape of bands")

        # here I build the x distances on the graph (in cartesian coordinates
        # if cartesian==True AND if the cell was set, otherwise in reciprocal
        # coordinates)
        try:
            kpoints = self.get_kpoints(cartesian=cartesian)
        except AttributeError:
            # this error is happening if cartesian==True and if no cell has been
            # set -> we switch to reciprocal coordinates to compute distances
            kpoints = self.get_kpoints()
        # I take advantage of the path to recognize discontinuities
        try:
            labels = self.labels
            labels_indices = [i[0] for i in labels]
        except (AttributeError, TypeError):
            labels = []
            labels_indices = []

        # since I can have discontinuous paths, I set on those points the distance to zero
        # as a result, where there are discontinuities in the path,
        # I have two consecutive points with the same x coordinate
        distances = [numpy.linalg.norm(kpoints[i] - kpoints[i - 1]) if not
        (i in labels_indices and i - 1 in labels_indices) else 0.
                     for i in range(1, len(kpoints))]
        x = [float(sum(distances[:i])) for i in range(len(distances) + 1)]

        # transform the index of the labels in the coordinates of x
        the_labels = [(x[i[0]], i[1]) for i in labels]

        if the_labels:
            new_labels = [list(the_labels[0])]
            # modify labels when in overlapping position
            j = 0
            for i in range(1, len(the_labels)):
                if new_labels[j][1] == 'G' and 'agr' in fileformat:
                    new_labels[j][1] = r"\xG"
                if the_labels[i][0] == the_labels[i - 1][0]:
                    new_labels[j][1] += "|" + the_labels[i][1]
                else:
                    new_labels.append(list(the_labels[i]))
                    j += 1
            if new_labels[j][1] == 'G' and 'agr' in fileformat:
                new_labels[j][1] = r"\xG"

        else:
            new_labels = []

        plot_info = {}
        plot_info['x'] = x
        plot_info['y'] = bands
        plot_info['labels'] = new_labels
        if path:
            plot_info['filename'] = path
            if fileformat == 'agr':
                plot_info['filename'] = os.path.splitext(path)[0] + '.dat'

        # generic info
        if comments:
            filetext = []
            # filetext.append("{}".format(get_file_header()))
            filetext.append("# Dumped from BandsData UUID={}"
                            .format(self.uuid))
            filetext.append("#\tpoints\tbands")
            filetext.append("#\t{}\t{}".format(*bands.shape))
            filetext.append("# \tlabel\tpoint")
            for l in new_labels:
                filetext.append("#\t{}\t{:.8f}".format(l[1], l[0]))
            filetext = get_file_header() + "#\n" + "\n".join(filetext) + "\n\n"
        else:
            filetext = ""

        try:
            newfiletext, extra_files = getattr(self, preparer_name)(plot_info, **kwargs)
            filetext += newfiletext
        except AttributeError:
            raise
        #            raise ValueError("Format {} is not valid".format(fileformat))

        if extra_files is not None:
            return filetext, extra_files
        else:
            return filetext

    def export(self, path, fileformat=None, overwrite=False, comments=True,
               cartesian=True, **kwargs):
        """
        Export the bands to a file.
        :param path: absolute path of the file to be created
        :param fileformat: format of the file created. If None, tries to use the
                     extension of path to understand the correct one.
        :param overwrite: if set to True, overwrites file found at path. Default=False
        :param comments: if True, append some extra informations at the
                beginning of the file.
        :param cartesian: if True, distances (for the x-axis) are computed in
                cartesian coordinates, otherwise they are computed in reciprocal
                coordinates. cartesian=True will fail if no cell has been set.
        :param **kwargs: additional parameters to be passed to the
                _prepare_[fileformat] method

        :note: this function will NOT produce nice plots if:
              - there is no path of kpoints, but a set of isolated points
              - the path is not continuous AND no labels are set
        """
        import os

        if not path:
            raise ValueError("Path not recognized")

        if os.path.exists(path) and not overwrite:
            raise OSError("A file was already found at {}".format(path))

        if fileformat is None:
            extension = os.path.splitext(path)[1].split('.')[1]
            if not extension:
                raise ValueError("Cannot recognized the fileformat from the "
                                 "extension")
            fileformat = extension
            if extension == 'dat':
                fileformat = 'dat_1'

        try:
            filetext, extra_files = self._exportstring(fileformat, path=path,
                                                       comments=comments, cartesian=cartesian,
                                                       **kwargs)
        except ValueError:
            extra_files = None
            filetext = self._exportstring(fileformat, path=path,
                                          comments=comments, cartesian=cartesian,
                                          **kwargs)

        if extra_files is not None:
            # write extra files
            for k, v in extra_files.iteritems():
                if os.path.exists(k) and not overwrite:
                    raise OSError("A file was already found: {}".format(k))
                else:
                    with open(k, "w") as f:
                        f.write(v)

        with open(path, 'w') as f:
            f.write(filetext)

    def get_export_formats(self):
        names = dir(self)
        return [i.split('_prepare_')[1] for i in names if i.startswith('_prepare_')]

    def _prepare_agr_batch(self, plot_info, **kwargs):
        """
        Prepare two files, data and batch, to be plot with xmgrace as:
        xmgrace -batch file.dat
        """
        if kwargs:
            raise TypeError("_prepare_agr_batch takes no keyword arguments")

        bands = plot_info['y']
        x = plot_info['x']
        labels = plot_info['labels']

        num_labels = len(labels)
        num_bands = bands.shape[1]

        # axis limits
        y_max_lim = bands.max()
        y_min_lim = bands.min()
        x_min_lim = min(x)  # this isn't a numpy array, but a list
        x_max_lim = max(x)

        # first prepare the xy coordinates of the sets
        raw_data, _ = self._prepare_dat_2(plot_info)

        # add the xy coordinates of the vertical lines
        for l in labels:
            new_block = ["{}\t{}".format(l[0], y_min_lim)]
            new_block.append("{}\t{}".format(l[0], y_max_lim))
            new_block.append("")
            raw_data += "\n".join(new_block)

        filexy_name = plot_info['filename']

        batch = []
        batch.append('READ XY "{}"'.format(filexy_name))

        # axis limits
        batch.append("world {}, {}, {}, {}".format(x_min_lim, y_min_lim, x_max_lim, y_max_lim))

        # axis label
        batch.append('yaxis label "Dispersion"')

        # axis ticks
        batch.append('xaxis  tick place both')
        batch.append('xaxis  tick spec type both')
        batch.append('xaxis  tick spec {}'.format(len(labels)))
        # set the name of the special points
        for i, l in enumerate(labels):
            batch.append("xaxis  tick major {}, {}".format(i, l[0]))
            batch.append('xaxis  ticklabel {}, "{}"'.format(i, l[1]))

        # minor graphical tweak
        batch.append("yaxis  tick minor ticks 3")
        batch.append("frame linewidth 2.0")

        # use helvetica fonts
        batch.append('map font 4 to "Helvetica", "Helvetica"')
        batch.append("yaxis  label font 4")
        batch.append("xaxis  label font 4")

        # set color and linewidths of bands
        for i in range(num_bands):
            batch.append("s{} line color 1".format(i))
            batch.append("s{} linewidth 1".format(i))

        # set color and linewidths of bands
        for i in range(num_bands, num_bands + num_labels):
            batch.append("s{} line color 1".format(i))
            batch.append("s{} linewidth 2".format(i))

        batch_data = "\n".join(batch) + "\n"
        extra_files = {"batch.dat": batch_data}

        return raw_data, extra_files

    def _prepare_dat_1(self, plot_info, **kwargs):
        """
        Write an N x M matrix. First column is the distance between kpoints,
        The other columns are the bands. Header contains number of kpoints and
        the number of bands (commented).
        """
        if kwargs:
            raise TypeError("_prepare_dat_1 takes no keyword arguments")

        bands = plot_info['y']
        x = plot_info['x']

        return_text = []

        for i in zip(x, bands):
            line = ["{:.8f}".format(i[0])] + ["{:.8f}".format(j) for j in i[1]]
            return_text.append("\t".join(line))

        return "\n".join(return_text) + '\n', None

    def _prepare_dat_2(self, plot_info, **kwargs):
        """
        Format suitable for gnuplot using blocks.
        Columns with x and y (path and band energy). Several blocks, separated
        by two empty lines, one per energy band.
        """
        if kwargs:
            raise TypeError("_prepare_dat_2 takes no keyword arguments")

        bands = plot_info['y']
        x = plot_info['x']

        return_text = []

        the_bands = numpy.transpose(bands)

        for b in the_bands:
            for i in zip(x, b):
                line = ["{:.8f}".format(i[0]), "{:.8f}".format(i[1])]
                return_text.append("\t".join(line))
            return_text.append("")
            return_text.append("")

        return "\n".join(return_text), None

    def _prepare_agr(self, plot_info, setnumber_offset=0, color_number=1,
                     legend="", title="", y_max_lim=None, y_min_lim=None,
                     y_origin=0., **kwargs):
        """
        Prepare an xmgrace agr file
        :param plot_info: a dictionary
        :param setnumber_offset: an offset to be applied to all set numbers
        (i.e. s0 is replaced by s[offset], s1 by s[offset+1], etc.)
        :param color_number: the color number for lines, symbols, error bars
        and filling (should be less than the parameter max_num_agr_colors
        defined below)
        :param legend: the legend (applied only to the first set)
        :param title: the title
        :param y_max_lim: the maximum on the y axis (if None, put the
            maximum of the bands)
        :param y_min_lim: the minimum on the y axis (if None, put the
            minimum of the bands)
        :param y_origin: the new origin of the y axis -> all bands are replaced
            by bands-y_origin
        """
        if kwargs:
            raise TypeError("_prepare_agr got unexpected keyword argument(s) {}"
                            "".format(kwargs.keys()))

        import math
        # load the x and y of every set
        if color_number > max_num_agr_colors:
            raise ValueError("Color number is too high (should be less than {})"
                             "".format(max_num_agr_colors))

        bands = plot_info['y'] - y_origin
        x = plot_info['x']
        the_bands = numpy.transpose(bands)
        labels = plot_info['labels']
        num_labels = len(labels)

        # axis limits
        if y_max_lim is None:
            y_max_lim = the_bands.max()
        if y_min_lim is None:
            y_min_lim = the_bands.min()
        x_min_lim = min(x)  # this isn't a numpy array, but a list
        x_max_lim = max(x)
        ytick_spacing = 10 ** int(math.log10((y_max_lim - y_min_lim)))

        # prepare xticks labels
        sx1 = ""
        for i, l in enumerate(labels):
            sx1 += agr_single_xtick_template.substitute(index=i,
                                                        coord=l[0],
                                                        name=l[1],
                                                        )
        xticks = agr_xticks_template.substitute(num_labels=num_labels,
                                                single_xtick_templates=sx1,
                                                )

        # build the arrays with the xy coordinates
        all_sets = []
        for b in the_bands:
            this_set = ""
            for i in zip(x, b):
                line = "{:.8f}".format(i[0]) + '\t' + "{:.8f}".format(i[1]) + "\n"
                this_set += line
            all_sets.append(this_set)

        set_descriptions = ""
        for i, this_set in enumerate(all_sets):
            width = str(2.0)
            set_descriptions += agr_set_description_template.substitute(
                set_number=i + setnumber_offset,
                linewidth=width,
                color_number=color_number,
                legend=legend if i == 0 else ""
            )

        units = self.units

        graphs = agr_graph_template.substitute(x_min_lim=x_min_lim,
                                               y_min_lim=y_min_lim,
                                               x_max_lim=x_max_lim,
                                               y_max_lim=y_max_lim,
                                               yaxislabel="Dispersion ({})".format(units),
                                               xticks_template=xticks,
                                               set_descriptions=set_descriptions,
                                               ytick_spacing=ytick_spacing,
                                               title=title,
                                               )
        sets = []
        for i, this_set in enumerate(all_sets):
            sets.append(agr_singleset_template.substitute(set_number=i + setnumber_offset,
                                                          xydata=this_set)
                        )
        the_sets = "&\n".join(sets)

        s = agr_template.substitute(graphs=graphs, sets=the_sets)

        return s, None

    def _prepare_json(self, plot_info, **kwargs):
        """
        Prepare a json file in a format compatible with the AiiDA band visualizer
        :param plot_info: a dictionary
        """
        import json

        if kwargs:
            raise TypeError("_prepare_json got unexpected keyword argument(s) {}"
                            "".format(kwargs.keys()))

        json_dict = {'label': self.label}

        bands = plot_info['y']
        labels = self.labels

        json_dict['path'] = []
        json_dict['paths'] = []

        try:
            _ = self.labels[1][1]
            for (position_from, label_from), (position_to, label_to) in zip(labels[:-1], labels[1:]):
                json_dict['path'].append([label_from, label_to])
                path_dict = {'length': position_to - position_from,
                             'from': label_from,
                             'to': label_to,
                             'values': bands[position_from:position_to + 1, :].transpose().tolist(),
                             }
                json_dict['paths'].append(path_dict)
        except (TypeError, IndexError):
            label_from = "0"
            label_to = "1"
            path_dict = {'length': bands.shape[0] - 1,
                         'from': label_from,
                         'to': label_to,
                         'values': bands.transpose().tolist(),
                         }
            json_dict['paths'].append(path_dict)
            json_dict['path'].append([label_from, label_to])

        return json.dumps(json_dict), None


max_num_agr_colors = 15

agr_template = Template(
    """
    # Grace project file
    #
    @version 50122
    @page size 792, 612
    @page scroll 5%
    @page inout 5%
    @link page off
    @map font 8 to "Courier", "Courier"
    @map font 10 to "Courier-Bold", "Courier-Bold"
    @map font 11 to "Courier-BoldOblique", "Courier-BoldOblique"
    @map font 9 to "Courier-Oblique", "Courier-Oblique"
    @map font 4 to "Helvetica", "Helvetica"
    @map font 6 to "Helvetica-Bold", "Helvetica-Bold"
    @map font 7 to "Helvetica-BoldOblique", "Helvetica-BoldOblique"
    @map font 5 to "Helvetica-Oblique", "Helvetica-Oblique"
    @map font 14 to "NimbusMonoL-BoldOblique", "NimbusMonoL-BoldOblique"
    @map font 15 to "NimbusMonoL-Regular", "NimbusMonoL-Regular"
    @map font 16 to "NimbusMonoL-RegularOblique", "NimbusMonoL-RegularOblique"
    @map font 17 to "NimbusRomanNo9L-Medium", "NimbusRomanNo9L-Medium"
    @map font 18 to "NimbusRomanNo9L-MediumItalic", "NimbusRomanNo9L-MediumItalic"
    @map font 19 to "NimbusRomanNo9L-Regular", "NimbusRomanNo9L-Regular"
    @map font 20 to "NimbusRomanNo9L-RegularItalic", "NimbusRomanNo9L-RegularItalic"
    @map font 21 to "NimbusSansL-Bold", "NimbusSansL-Bold"
    @map font 22 to "NimbusSansL-BoldCondensed", "NimbusSansL-BoldCondensed"
    @map font 23 to "NimbusSansL-BoldCondensedItalic", "NimbusSansL-BoldCondensedItalic"
    @map font 24 to "NimbusSansL-BoldItalic", "NimbusSansL-BoldItalic"
    @map font 25 to "NimbusSansL-Regular", "NimbusSansL-Regular"
    @map font 26 to "NimbusSansL-RegularCondensed", "NimbusSansL-RegularCondensed"
    @map font 27 to "NimbusSansL-RegularCondensedItalic", "NimbusSansL-RegularCondensedItalic"
    @map font 28 to "NimbusSansL-RegularItalic", "NimbusSansL-RegularItalic"
    @map font 29 to "StandardSymbolsL-Regular", "StandardSymbolsL-Regular"
    @map font 12 to "Symbol", "Symbol"
    @map font 31 to "Symbol-Regular", "Symbol-Regular"
    @map font 2 to "Times-Bold", "Times-Bold"
    @map font 3 to "Times-BoldItalic", "Times-BoldItalic"
    @map font 1 to "Times-Italic", "Times-Italic"
    @map font 0 to "Times-Roman", "Times-Roman"
    @map font 36 to "URWBookmanL-DemiBold", "URWBookmanL-DemiBold"
    @map font 37 to "URWBookmanL-DemiBoldItalic", "URWBookmanL-DemiBoldItalic"
    @map font 38 to "URWBookmanL-Light", "URWBookmanL-Light"
    @map font 39 to "URWBookmanL-LightItalic", "URWBookmanL-LightItalic"
    @map font 40 to "URWChanceryL-MediumItalic", "URWChanceryL-MediumItalic"
    @map font 41 to "URWGothicL-Book", "URWGothicL-Book"
    @map font 42 to "URWGothicL-BookOblique", "URWGothicL-BookOblique"
    @map font 43 to "URWGothicL-Demi", "URWGothicL-Demi"
    @map font 44 to "URWGothicL-DemiOblique", "URWGothicL-DemiOblique"
    @map font 45 to "URWPalladioL-Bold", "URWPalladioL-Bold"
    @map font 46 to "URWPalladioL-BoldItalic", "URWPalladioL-BoldItalic"
    @map font 47 to "URWPalladioL-Italic", "URWPalladioL-Italic"
    @map font 48 to "URWPalladioL-Roman", "URWPalladioL-Roman"
    @map font 13 to "ZapfDingbats", "ZapfDingbats"
    @map color 0 to (255, 255, 255), "white"
    @map color 1 to (0, 0, 0), "black"
    @map color 2 to (255, 0, 0), "red"
    @map color 3 to (0, 255, 0), "green"
    @map color 4 to (0, 0, 255), "blue"
    @map color 5 to (255, 215, 0), "yellow"
    @map color 6 to (188, 143, 143), "brown"
    @map color 7 to (220, 220, 220), "grey"
    @map color 8 to (148, 0, 211), "violet"
    @map color 9 to (0, 255, 255), "cyan"
    @map color 10 to (255, 0, 255), "magenta"
    @map color 11 to (255, 165, 0), "orange"
    @map color 12 to (114, 33, 188), "indigo"
    @map color 13 to (103, 7, 72), "maroon"
    @map color 14 to (64, 224, 208), "turquoise"
    @map color 15 to (0, 139, 0), "green4"
    @reference date 0
    @date wrap off
    @date wrap year 1950
    @default linewidth 1.0
    @default linestyle 1
    @default color 1
    @default pattern 1
    @default font 0
    @default char size 1.000000
    @default symbol size 1.000000
    @default sformat "%.8g"
    @background color 0
    @page background fill on
    @timestamp off
    @timestamp 0.03, 0.03
    @timestamp color 1
    @timestamp rot 0
    @timestamp font 0
    @timestamp char size 1.000000
    @timestamp def "Wed Jul 30 16:44:34 2014"
    @r0 off
    @link r0 to g0
    @r0 type above
    @r0 linestyle 1
    @r0 linewidth 1.0
    @r0 color 1
    @r0 line 0, 0, 0, 0
    @r1 off
    @link r1 to g0
    @r1 type above
    @r1 linestyle 1
    @r1 linewidth 1.0
    @r1 color 1
    @r1 line 0, 0, 0, 0
    @r2 off
    @link r2 to g0
    @r2 type above
    @r2 linestyle 1
    @r2 linewidth 1.0
    @r2 color 1
    @r2 line 0, 0, 0, 0
    @r3 off
    @link r3 to g0
    @r3 type above
    @r3 linestyle 1
    @r3 linewidth 1.0
    @r3 color 1
    @r3 line 0, 0, 0, 0
    @r4 off
    @link r4 to g0
    @r4 type above
    @r4 linestyle 1
    @r4 linewidth 1.0
    @r4 color 1
    @r4 line 0, 0, 0, 0
    $graphs
    $sets
    """
)

agr_xticks_template = Template(
    """
    @    xaxis  tick spec $num_labels
    $single_xtick_templates
    """)

agr_single_xtick_template = Template(
    """
    @    xaxis  tick major $index, $coord
    @    xaxis  ticklabel $index, "$name"
    """)

agr_graph_template = Template(
    """
    @g0 on
    @g0 hidden false
    @g0 type XY
    @g0 stacked false
    @g0 bar hgap 0.000000
    @g0 fixedpoint off
    @g0 fixedpoint type 0
    @g0 fixedpoint xy 0.000000, 0.000000
    @g0 fixedpoint format general general
    @g0 fixedpoint prec 6, 6
    @with g0
    @    world $x_min_lim, $y_min_lim, $x_max_lim, $y_max_lim
    @    stack world 0, 0, 0, 0
    @    znorm 1
    @    view 0.150000, 0.150000, 1.150000, 0.850000
    @    title "$title"
    @    title font 0
    @    title size 1.500000
    @    title color 1
    @    subtitle ""
    @    subtitle font 0
    @    subtitle size 1.000000
    @    subtitle color 1
    @    xaxes scale Normal
    @    yaxes scale Normal
    @    xaxes invert off
    @    yaxes invert off
    @    xaxis  on
    @    xaxis  type zero false
    @    xaxis  offset 0.000000 , 0.000000
    @    xaxis  bar on
    @    xaxis  bar color 1
    @    xaxis  bar linestyle 1
    @    xaxis  bar linewidth 1.0
    @    xaxis  label ""
    @    xaxis  label layout para
    @    xaxis  label place auto
    @    xaxis  label char size 1.000000
    @    xaxis  label font 4
    @    xaxis  label color 1
    @    xaxis  label place normal
    @    xaxis  tick on
    @    xaxis  tick major 5
    @    xaxis  tick minor ticks 0
    @    xaxis  tick default 6
    @    xaxis  tick place rounded true
    @    xaxis  tick in
    @    xaxis  tick major size 1.000000
    @    xaxis  tick major color 1
    @    xaxis  tick major linewidth 1.0
    @    xaxis  tick major linestyle 1
    @    xaxis  tick major grid on
    @    xaxis  tick minor color 1
    @    xaxis  tick minor linewidth 1.0
    @    xaxis  tick minor linestyle 1
    @    xaxis  tick minor grid off
    @    xaxis  tick minor size 0.500000
    @    xaxis  ticklabel on
    @    xaxis  ticklabel format general
    @    xaxis  ticklabel prec 5
    @    xaxis  ticklabel formula ""
    @    xaxis  ticklabel append ""
    @    xaxis  ticklabel prepend ""
    @    xaxis  ticklabel angle 0
    @    xaxis  ticklabel skip 0
    @    xaxis  ticklabel stagger 0
    @    xaxis  ticklabel place normal
    @    xaxis  ticklabel offset auto
    @    xaxis  ticklabel offset 0.000000 , 0.010000
    @    xaxis  ticklabel start type auto
    @    xaxis  ticklabel start 0.000000
    @    xaxis  ticklabel stop type auto
    @    xaxis  ticklabel stop 0.000000
    @    xaxis  ticklabel char size 1.500000
    @    xaxis  ticklabel font 4
    @    xaxis  ticklabel color 1
    @    xaxis  tick place both
    @    xaxis  tick spec type both
    $xticks_template
    @    yaxis  on
    @    yaxis  type zero false
    @    yaxis  offset 0.000000 , 0.000000
    @    yaxis  bar on
    @    yaxis  bar color 1
    @    yaxis  bar linestyle 1
    @    yaxis  bar linewidth 1.0
    @    yaxis  label "$yaxislabel"
    @    yaxis  label layout para
    @    yaxis  label place auto
    @    yaxis  label char size 1.500000
    @    yaxis  label font 4
    @    yaxis  label color 1
    @    yaxis  label place normal
    @    yaxis  tick on
    @    yaxis  tick major $ytick_spacing
    @    yaxis  tick minor ticks 1
    @    yaxis  tick default 6
    @    yaxis  tick place rounded true
    @    yaxis  tick in
    @    yaxis  tick major size 1.000000
    @    yaxis  tick major color 1
    @    yaxis  tick major linewidth 1.0
    @    yaxis  tick major linestyle 1
    @    yaxis  tick major grid off
    @    yaxis  tick minor color 1
    @    yaxis  tick minor linewidth 1.0
    @    yaxis  tick minor linestyle 1
    @    yaxis  tick minor grid off
    @    yaxis  tick minor size 0.500000
    @    yaxis  ticklabel on
    @    yaxis  ticklabel format general
    @    yaxis  ticklabel prec 5
    @    yaxis  ticklabel formula ""
    @    yaxis  ticklabel append ""
    @    yaxis  ticklabel prepend ""
    @    yaxis  ticklabel angle 0
    @    yaxis  ticklabel skip 0
    @    yaxis  ticklabel stagger 0
    @    yaxis  ticklabel place normal
    @    yaxis  ticklabel offset auto
    @    yaxis  ticklabel offset 0.000000 , 0.010000
    @    yaxis  ticklabel start type auto
    @    yaxis  ticklabel start 0.000000
    @    yaxis  ticklabel stop type auto
    @    yaxis  ticklabel stop 0.000000
    @    yaxis  ticklabel char size 1.250000
    @    yaxis  ticklabel font 4
    @    yaxis  ticklabel color 1
    @    yaxis  tick place both
    @    yaxis  tick spec type none
    @    altxaxis  off
    @    altyaxis  off
    @    legend on
    @    legend loctype view
    @    legend 0.85, 0.8
    @    legend box color 1
    @    legend box pattern 1
    @    legend box linewidth 1.0
    @    legend box linestyle 1
    @    legend box fill color 0
    @    legend box fill pattern 1
    @    legend font 0
    @    legend char size 1.000000
    @    legend color 1
    @    legend length 4
    @    legend vgap 1
    @    legend hgap 1
    @    legend invert false
    @    frame type 0
    @    frame linestyle 1
    @    frame linewidth 1.0
    @    frame color 1
    @    frame pattern 1
    @    frame background color 0
    @    frame background pattern 0
    $set_descriptions
    """
)

agr_set_description_template = Template(
    """
    @    s$set_number hidden false
    @    s$set_number type xy
    @    s$set_number symbol 0
    @    s$set_number symbol size 1.000000
    @    s$set_number symbol color $color_number
    @    s$set_number symbol pattern 1
    @    s$set_number symbol fill color $color_number
    @    s$set_number symbol fill pattern 0
    @    s$set_number symbol linewidth 1.0
    @    s$set_number symbol linestyle 1
    @    s$set_number symbol char 65
    @    s$set_number symbol char font 0
    @    s$set_number symbol skip 0
    @    s$set_number line type 1
    @    s$set_number line linestyle 1
    @    s$set_number line linewidth $linewidth
    @    s$set_number line color $color_number
    @    s$set_number line pattern 1
    @    s$set_number baseline type 0
    @    s$set_number baseline off
    @    s$set_number dropline off
    @    s$set_number fill type 0
    @    s$set_number fill rule 0
    @    s$set_number fill color $color_number
    @    s$set_number fill pattern 1
    @    s$set_number avalue off
    @    s$set_number avalue type 2
    @    s$set_number avalue char size 1.000000
    @    s$set_number avalue font 0
    @    s$set_number avalue color 1
    @    s$set_number avalue rot 0
    @    s$set_number avalue format general
    @    s$set_number avalue prec 3
    @    s$set_number avalue prepend ""
    @    s$set_number avalue append ""
    @    s$set_number avalue offset 0.000000 , 0.000000
    @    s$set_number errorbar on
    @    s$set_number errorbar place both
    @    s$set_number errorbar color $color_number
    @    s$set_number errorbar pattern 1
    @    s$set_number errorbar size 1.000000
    @    s$set_number errorbar linewidth 1.0
    @    s$set_number errorbar linestyle 1
    @    s$set_number errorbar riser linewidth 1.0
    @    s$set_number errorbar riser linestyle 1
    @    s$set_number errorbar riser clip off
    @    s$set_number errorbar riser clip length 0.100000
    @    s$set_number comment "Cols 1:2"
    @    s$set_number legend "$legend"
    """
)

agr_singleset_template = Template(
    """
    @target G0.S$set_number
    @type xy
    $xydata
    """)



# -*- coding: utf-8 -*-

from sqlalchemy import create_engine
from sqlalchemy.exc import SQLAlchemyError


__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__authors__ = "The AiiDA team."
__version__ = "0.7.0.1"

def create_connection():
    engine = create_engine(
        'postgresql://aiida:aiida_db@localhost:5433/aiida_separate'
    )

    return engine.connect()


def migrate_denormalized(conn):
    # trans = conn.begin()
    # try:
    #     result = conn.execute("SELECT DISTINCT type FROM db_dbnode;")
    #
    #     values = set([e["type"] for e in result])
    #
    #     conn.execute("DROP INDEX IF EXISTS db_dbnode_type;")
    #     conn.execute("CREATE TABLE db_dbnode_type (id serial CONSTRAINT db_dbnode_type_pk PRIMARY KEY, name varchar(255) not null);")
    #
    #     conn.execute("INSERT INTO db_dbnode_type (name) VALUES {};".format(
    #         ", ".join(["('{}')".format(e) for e in values])
    #     ))
    #
    #     trans.commit()
    # except SQLAlchemyError as e:
    #     print("Error: {}".format(e))
    #     trans.rollback()

    trans = conn.begin()
    try:
        # result = conn.execute("SELECT id, name FROM db_dbnode_type;")
        # values = dict((v["name"], v["id"]) for v in result)
        #
        conn.execute(
            "ALTER TABLE db_dbnode ADD COLUMN type_id integer"
        )
        conn.execute(
            "ALTER TABLE db_dbnode ADD FOREIGN KEY (type_id) REFERENCES db_dbnode_type(id);"
        )

        conn.execute(
            "UPDATE db_dbnode SET type_id = db_dbnode_type.id "
            "FROM db_dbnode_type WHERE db_dbnode_type.name = db_dbnode.type"
        )

        # conn.execute(
        #     "ALTER TABLE db_dbnode DROP COLUMN type"
        # )
        trans.commit()
    except SQLAlchemyError as e:
        print("Error: {}".format(e))
        trans.rollback()


